%{
/* C declarations */

#define YYDEBUG 1
#define YYERROR_VERBOSE 1

#include <stdio.h>
extern void yyStartFile(FILE *fp, const char *name);
extern void yyerror(const char *msg);
extern int yylex();

#include "expr.hh"
#include "dimdef.hh"
#include "casevalue.hh"
#include "binding.hh"

%}
/* BISON declarations */

/* Expect 5 shift/reduce conflicts. */
%expect 5
/* The conflicts come from:
 *     "anylabeldef :"  vs "subcall_implicit :" (MAKES 2?)
 *        xx: y   can be interpreted as that xx is a label or a subname
 *                qbasic takes xx as a label, we'll obey
 *
 *     shortifblock
 *          IF rule1 THEN IF rule2 THEN st1 ELSE st2
 *             can be translated as:
 *                   - st2 if not rule1
 *                   - st2 if not rule2
 *             it makes up three shift/reduce conflicts:
 *                   what are they? todo.
 *                    
 */

%expect-rr 0

%start program

%union {
    /* non-POD objects are not allowed in union, thus a pointer must be used */
    class expr           *expr;
    class casevalues     *casevalues;
    class expressions    *exprlist;
    std::string          *string;
    class dimindexdef    *dimindexdef;
    class statementchain *function;
    struct {} token;
}

%type<expr>        expr expr_source builtinfunc
%type<expr>        numexpr boolexpr stringexpr string_or_num_expr dummy_expr 
%type<expr>        var_target_name var_target forstep
%type<expr>        var_or_const_or_function_name var_or_const_or_functioncall

%type<expr>        NUMBER FILENUM autofilenum

%type<casevalues>  casevalue1 casevalues2

%type<exprlist>    subcallparams0 subcallparams2 indexlist

%type<string>      IDENTIFIER
%type<string>      STRINGCONST

%type<dimindexdef> dimindexdef1 dimindexdefs2

%type<function>    shortstatement1 shortstatements0 shortif shortifblock
%type<function>    shortelseblock shortelseblockB shortdoloop
%type<function>    shortfornext shortselect shortcaselist
%type<function>    longstatement1 longstatements0 longstructure longif
%type<function>    longelseiflist longelse longdoloop longfornext
%type<function>    longselect longcaselist anylabeldef
%type<function>    simplestatement
%type<function>    subcall subcall_explicit subcall_implicit

%type<string>      anylabeltarget anylabellist

/* This doesn't work */
/*%left ELSE
%right COLONCHAR*/

/* Operator precedence - least precedence group */
%left IMP EQV   /* 1101, 1001 */
%left XOR OR    /* 0110, 0111 */
%left AND       /* 0001 */

%right NOT  /* 0=1 1=0 */

/* Operator precedence - middle precedence group */
%left '<' '>' '=' LESSOREQUAL GREATEROREQUAL NOTEQUAL

/* Operator precedence - top precedence group */
%left '+' '-'
%left MOD
%left '\\'
%left '*' '/'

%right NEGATIONPRIORITY
%right '^'

%token CRLF COLONCHAR

%token IF THEN ELSE ELSEIF ENDIF

%token SELECT CASE ENDSELECT IS

%token DO LOOP EXITDO
%token DOWHILE DOUNTIL LOOPWHILE LOOPUNTIL

%token SUB ENDSUB EXITSUB DECLARESUB
%token FUNCTION ENDFUNCTION EXITFUNCTION DECLAREFUNCTION
%token CALL

%token FOR TO STEP NEXT EXITFOR
%token GOTO GOSUB RETURN
%token END
%token LET

%token NUMBER STRINGCONST
%token IDENTIFIER

%token PRINT PRINT_F USING FILENUM LPRINT OPEN
%token OUTPUT INPUT APPEND RANDOM BINARY WRITE ACCESS LOCK CLOSE

%token PUT GET SCREEN VIEW PALETTE PSET PRESET
%token LINE CIRCLE PAINT DRAW COLOR CLS

%token ON_PLAY ON PLAY STOP OFF POKE SYSTEM LOCATE ERROR RESUME OUT WAIT
%token DEFSEG

	/* Functions without parameters*/
%token INKEY_S
%token CSRLIN
%token FREEFILE
%token DATE_S
%token TIME_S
%token TIMER
%token ERL
%token ERR
%token RND
        /* Functions taking strings */
%token ASC
%token LEN
%token CVD
%token CVI
%token CVL
%token CVS
%token LTRIM_S
%token RTRIM_S
%token LCASE_S
%token UCASE_S
%token VAL
        /* Functions with mixed parameters */
%token ENVIRON_S
%token FILEATTR
%token FRE
%token INSTR
%token MID_S
%token LEFT_S
%token RIGHT_S
        /* Functions taking numbers */
%token ABS
%token ATN
%token CDBL
%token CINT
%token CHR_S
%token CLNG
%token COS
%token CSNG
%token EXP
%token FIX
%token HEX_S
%token INP
%token INPUT_S
%token INT
%token LOC
%token LOF
%token LOG
%token LPOS
%token MKD_S
%token MKI_S
%token MKL_S
%token MKS_S
%token OCT_S
%token POINT
%token PEEK
%token POS
%token SPC
%token SQR
%token STR_S
%token STRING_S
%token TAB
%token TAN
%token SPACE_S
%token LBOUND
%token UBOUND
%token VARPTR
%token VARPTR_S
%token VARSEG


%token DATA READ RESTORE

%token INTEGER SINGLE DOUBLE STRING LONG AS
%token TYPE ENDTYPE
%token DIM REDIM SHARED CONST OPTION_BASE STATIC

%token UNKNOWN_CHARACTER

%%
/* Grammar rules */

program:
		definitions0
;

definitions0:
		definitions0 definition1
	|	/* empty */
;

definition1:
		sub
	|	function
	|	declaresub
	|	declarefunction
	|	typedef
	|	constdef
	|	OPTION_BASE int_const separator1
		{
		}
	|	longstatement1 separator1
	|	shortif CRLF
	|	anylabeldef
	|	separator1
;

sub:
		SUB IDENTIFIER sub_params_decl sub_static_flag separator1
		    longstatements0 ENDSUB CRLF
		{
			/* define identifier as a sub name */
			if(IsDefinedAsNonSub(*$2)) { YYABORT; }
			
			/* define the given global sub */
		}
;
function:
		FUNCTION IDENTIFIER sub_params_decl sub_static_flag separator1
		         longstatements0 ENDFUNCTION CRLF
		{
			/* define identifier as a function name */
			if(IsDefinedAsNonFunc(*$2)) { YYABORT; }
			
			/* define the given global function */
		}
;
declaresub:
		DECLARESUB IDENTIFIER sub_params_decl CRLF
		{
			/* define identifier as a sub name */
			if(IsDefinedAsSubFuncVarConst(*$2)) { YYABORT; }
			
			/* introduce the given global sub to namespace */
		}
;
declarefunction:
		DECLAREFUNCTION IDENTIFIER sub_params_decl CRLF
		{
			/* define identifier as a function name */
			if(IsDefinedAsSubFuncVarConst(*$2)) { YYABORT; }
			
			/* introduce the given global function to namespace */
		}
;

typedef:
		TYPE IDENTIFIER CRLF typeelements2 ENDTYPE CRLF
		{
			/* define identifier as a type */
			if(IsType(*$2)) { YYABORT; }
		}
;


constdef:
		CONST IDENTIFIER '=' expr CRLF
		{
			if(IsDefinedAsSubFuncVarConst(*$2)) { YYABORT; }
			/* TODO: define the const as an "expr" */
		}
;

/* This does not terminate with anything */
shortstatement1:
		simplestatement
	|	shortdoloop
	|	shortfornext
	|	shortselect
	|	error
;
shortstatements0: /* Used where short statements are required */
                  /* May not contain shortifs */
                  /* Ends with COLONCHARs */
		shortstatements0 shortstatement1 COLONCHAR
	|	shortstatements0 COLONCHAR { $$ = $1 }
	|	/* empty */ { }
;

shortif:
		IF boolexpr THEN shortifblock { }
;

/* shortifblock doesn't end with COLONCHAR */
shortifblock:
		shortstatement1 COLONCHAR shortifblock
	|	shortstatement1 ELSE shortelseblock /* conflict 2 */
	|	shortstatement1
	|	shortif ELSE shortelseblock /* conflict 3 */
	|	shortif
	|	COLONCHAR shortifblock { $$ = $2 }
	|	ELSE shortelseblock /* no conflict?? */ { }
	|	int_const ELSE shortelseblock /* conflict 1 */ { }
	|	int_const { }
;

shortelseblock:
		int_const
	|	shortelseblockB
	|	/* empty */ { }
;
shortelseblockB:
		shortstatement1 COLONCHAR shortelseblockB
	|	shortstatement1
	|	shortif
;

shortdoloop:
		DO COLONCHAR shortstatements0 LOOP
	|	DOWHILE boolexpr COLONCHAR shortstatements0 LOOP
	|	DOUNTIL boolexpr COLONCHAR shortstatements0 LOOP
	|	DO COLONCHAR shortstatements0 LOOPWHILE boolexpr
	|	DO COLONCHAR shortstatements0 LOOPUNTIL boolexpr
;
shortfornext:
		FOR var_target '=' numexpr TO numexpr forstep COLONCHAR shortstatements0 NEXT
;
shortselect:
		SELECT CASE expr COLONCHAR shortcaselist ENDSELECT
;
shortcaselist:
		shortcaselist CASE casevalues2 COLONCHAR shortstatements0
	|	/* empty */ { }
;

longstatement1:
		longstructure
	|	simplestatement
	|	error
;
longstatements0:
                /* Ends with any separator */
		longstatements0 longstatement1 separator1
	|	longstatements0 shortif CRLF
	|	longstatements0 anylabeldef
	|	longstatements0 separator1 { $$ = $1 }
	|	/* empty */ { }
;

longstructure: /* None of these is expected to be terminated with CRLF */
		longif
	|	longdoloop
	|	longfornext
	|	longselect
;
longif:
		IF boolexpr THEN CRLF longstatements0 longelseiflist longelse ENDIF
;
longelseiflist:
		longelseiflist ELSEIF boolexpr THEN longstatements0
	|	/* empty */ { }
;
longelse:
		ELSE longstatements0
	|	/* empty */ { }
;
longdoloop:
		DO separator1 longstatements0 LOOP
	|	DOWHILE boolexpr separator1 longstatements0 LOOP
	|	DOUNTIL boolexpr separator1 longstatements0 LOOP
	|	DO separator1 longstatements0 LOOPWHILE boolexpr
	|	DO separator1 longstatements0 LOOPUNTIL boolexpr
;
longfornext:
		FOR var_target '=' numexpr TO numexpr forstep separator1 longstatements0 NEXT
;
longselect:
		SELECT CASE expr separators2 longcaselist ENDSELECT
;
longcaselist:
		longcaselist CASE casevalues2 separator1 longstatements0
	|	/* empty */ { }
;


forstep:
		STEP numexpr
	|	/* empty */ { }
;
casevalue1:
		ELSE                    { $$ = new casevalues(ELSE);   }
	|	IS '>' expr             { $$ = new casevalues('>', $3) }
	|	IS '<' expr             { $$ = new casevalues('<', $3) }
	|	IS '=' expr             { $$ = new casevalues('=', $3) }
	|	expr                    { $$ = new casevalues('=', $1) }
	|	IS LESSOREQUAL expr     { $$ = new casevalues(LESSOREQUAL, $3) }
	|	IS GREATEROREQUAL expr  { $$ = new casevalues(GREATEROREQUAL, $3) }
	|	IS NOTEQUAL expr        { $$ = new casevalues(NOTEQUAL, $3) }
	|	expr TO expr            { $$ = new casevalues($1, $3) }
;
casevalues2:
		casevalues2 ',' casevalue1  { $1->set_as($3); $3=0; $$ = $1 }
	|	casevalue1                  { $$ = $1 }
;

separator1:
		COLONCHAR
	|	CRLF
;
/* unused
separators0:
		separators2
	|	-empty-
;
*/
separators2:
		separators2 separator1
	|	separator1
;

/* a label statement */
anylabeldef:
		int_const
	|	IDENTIFIER COLONCHAR
		{
			/* automatically define as label */
		}
;


anylabeltarget:
		IDENTIFIER
		{
			/* automatically define as label */
		}
	|	int_const
;
anylabellist:
		anylabellist ',' anylabeltarget
	|	anylabeltarget
;


/* Something that is not a control structure */
/* This does not terminate with anything */
simplestatement:
		LET assignment
	|	assignment
	|	RETURN anylabeltarget
	|	RETURN
	|	EXITDO
	|	EXITFOR
	|	EXITSUB
	|	EXITFUNCTION
	|	subcall
	|	dim
	|	GOTO anylabeltarget
	|	GOSUB anylabeltarget
	|	RESTORE anylabeltarget
	|	RESTORE
	|	ON_PLAY '(' numexpr ')' GOSUB anylabeltarget
	|	ON numexpr GOSUB anylabellist
	|	ON numexpr GOTO anylabellist
	|	ON ERROR GOTO anylabeltarget
	|	RESUME anylabeltarget
	|	RESUME NEXT
	|	RESUME
	|	PRINT_F autofilenum ',' printspec
	|	PRINT printspec
	|	LPRINT printspec
	|	POKE numexpr ',' numexpr
	|	PLAY STOP
	|	PLAY OFF
	|	PLAY ON
	|	PLAY playparams2
	|	READ readparams2
	|	DATA dataparams0
		 /* note: put DATA to global level, not in block */
	|	OUT numexpr ',' numexpr
	|	WAIT numexpr ',' numexpr ',' numexpr
	|	WAIT numexpr ',' numexpr
	|	SYSTEM stringexpr
	|	SYSTEM
	|	END
	|	STOP
	|	graphics
	|	open
	|	CLOSE closeparams2
	|	CLOSE /* close all files */
	|	DEFSEG '=' numexpr
	|	DEFSEG
;

dataparams0:
		dataparams0 ',' STRINGCONST
	|	dataparams0 ','
	|	STRINGCONST
	|	/* empty */
;
readparams2:
		readparams2 ',' var_target
	|	var_target
;

closeparams2:
		closeparams2 ',' autofilenum
	|	autofilenum
;

playparams2:
		playparams2 ',' stringexpr
	|	stringexpr
;

graphics:
		LOCATE optional_numexpr5
	                 /* row, col, cursor_visible, cursor_start, cursor_end */
	|	SCREEN optional_numexpr4
	                 /* mode, colorswitch, apage, vpage */
	|	GET coordinate '-' coordinate ',' var_target
	|	PUT coordinate ',' var_target optional_puttype
	|	PALETTE numexpr ',' numexpr
	|	PALETTE USING var_target
	|	PALETTE
	|	VIEW PRINT numexpr TO numexpr
	|	VIEW PRINT
	|	VIEW SCREEN plain_coordinate '-' plain_coordinate optional_numparams2
	|	VIEW        plain_coordinate '-' plain_coordinate optional_numparams2
	|	PSET coordinate optional_numparam
	|	PRESET coordinate optional_numparam
	|	LINE optional_coordinate '-' coordinate ',' optional_numexpr ',' optional_linetype optional_numparam
	|	LINE optional_coordinate '-' coordinate optional_numparam
	|	CIRCLE coordinate ',' numexpr optional_numparams4
	                 /* radius, color, start, end, aspect */
	|	PAINT coordinate ',' string_or_num_expr optional_numparams2
	|	PAINT coordinate ','                    optional_numparams2
	|       PAINT coordinate
	                 /* pattern/color, bordercolor, backgroundcolor */
	|	DRAW stringexpr
	|	COLOR optional_numexpr3
	|	CLS optional_numexpr
;

put_type:
		PSET
	|	PRESET
	|	AND
	|	OR
	|	XOR
;
optional_puttype:
		',' put_type
	|	/* empty - XOR assumed */
;
optional_linetype:
		IDENTIFIER /* B or BF */
	|	/* empty */
;

optional_numparam:
		',' optional_numexpr
	|	/* empty - assume "-1" */
;
optional_numparams2:
		',' optional_numexpr optional_numparam
	|	/* empty - assume "-1" */
;
optional_numparams3:
		',' optional_numexpr optional_numparams2
	|	/* empty - assume "-1" */
;
optional_numparams4:
		',' optional_numexpr optional_numparams3
	|	/* empty - assume "-1" */
;

optional_numexpr:
		numexpr
	|	/* empty - assume "-1" */
;
/*optional_numexpr2: optional_numexpr optional_numparam;*/
optional_numexpr3: optional_numexpr optional_numparams2;
optional_numexpr4: optional_numexpr optional_numparams3;
optional_numexpr5: optional_numexpr optional_numparams4;



open:		OPEN stringexpr /* filename */
                     open_formode open_accessmode open_lockmode
                     AS autofilenum
                     open_lenmode
        |	OPEN stringexpr /* mode */
                     ',' autofilenum
                     ',' stringexpr /* filename */
                     open_reclenmode
;

open_formode:
		FOR OUTPUT
	|	FOR INPUT
	|	FOR APPEND
	|	FOR RANDOM
	|	FOR BINARY
	|	/* empty - RANDOM assumed */
;
open_accessmode:
		ACCESS READ WRITE
	|	ACCESS READ
	|	ACCESS WRITE
	|	/* empty */
;
open_lockmode:
		SHARED
	|	LOCK READ WRITE
	|	LOCK READ
	|	LOCK WRITE
	|	/* empty */
;
open_lenmode:
		LEN '=' numexpr
	|	/* empty */
;
open_reclenmode:
		',' numexpr
	|	/* empty */
;

assignment:
		var_target '=' expr
;

subcall:
		subcall_explicit
	|	subcall_implicit
;
subcall_explicit:
		CALL IDENTIFIER '(' subcallparams2 ')'
		{
			if(IsDefinedAsNonSub(*$2)) { YYABORT; }
		}
	|	CALL IDENTIFIER
		{
			if(IsDefinedAsNonSub(*$2)) { YYABORT; }
		}
;
subcall_implicit:
		IDENTIFIER subcallparams0  /* Calling a SUB */
		{
			if(IsDefinedAsNonSub(*$1)) { YYABORT; }
		}
;

autofilenum:
		FILENUM
	|	numexpr
;

subcallparams0:
		subcallparams2           { $$ = $1 }
	|	/* empty */              { $$ = new expressions }
;
subcallparams2:
		subcallparams2 ',' expr  { $1->push_back($3); $$ = $1     }
	|	expr                     { $$ = new expressions($1)       }
;


sub_params_decl:
		'(' paramdeclarations2 ')'
	|	'(' ')'
	|	/* empty */
;
sub_static_flag:
		STATIC
	|	/* empty */
;


paramdeclaration1:
		IDENTIFIER AS type
		{
			if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; }
		}
	|	IDENTIFIER
		{
			if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; }
		}
;
paramdeclarations2:
		paramdeclarations2 ',' paramdeclaration1
	|	paramdeclaration1
;


typeelements2:
		typeelements2 typeelement1
	|	typeelement1
;
typeelement1:
		typeelementdef CRLF
	|	CRLF /* empty line is ok in typedef */
;
typeelementdef:
		typeelementdef ','
		IDENTIFIER AS type /* extension */
		{
			/* Element */
		}
	|	IDENTIFIER AS type
		{
			/* Element */
		}
;




printspec:
		USING stringexpr ';' printparams0
	|	printparams0
;
printparams0:
		printparams0 ';' expr
	|	printparams0 ';'
	|	printparams0 ',' expr
	|	printparams0 ','
	|	expr
	|	/* empty */
;

dim:
		DIM SHARED dimdefs2
	|	DIM dimdefs2
	|	REDIM redimdefs2
;
dimdefs2:
		dimdefs2 ',' dimdef1
	|	dimdef1
;
dimdef1:
		IDENTIFIER '(' dimindexdefs2 ')' AS type
		{
			if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; }
		}
	|	IDENTIFIER '(' dimindexdefs2 ')'
		{
			if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; }
		}
	|	IDENTIFIER AS type
		{
			if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; }
		}
	|	IDENTIFIER
		{
			if(IsDefinedAsSubFuncVarConst(*$1)) { YYABORT; }
		}
;
redimdefs2:
		redimdefs2 ',' redimdef1
	|	redimdef1
;
redimdef1:
		IDENTIFIER '(' dimindexdefs2 ')' AS type
		{
			if(!IsDefinedAsRedimmable(*$1)) { YYABORT; }
		}
	|	IDENTIFIER '(' dimindexdefs2 ')'
		{
			if(!IsDefinedAsRedimmable(*$1)) { YYABORT; }
		}
;
dimindexdefs2:
		dimindexdefs2 ',' dimindexdef1 { $1->set_as($3); $3=0; $$ = $1 }
	|	dimindexdef1                   { $$ = $1              }
;
dimindexdef1:
		numexpr TO numexpr             { $$ = new dimindexdef($1, $3); }
	|	numexpr                        { $$ = new dimindexdef($1);     }
;

expr:
		expr IMP expr                    { $$ = new expr_binary($1, IMP, $3); }
	|	expr EQV expr                    { $$ = new expr_binary($1, EQV, $3); }
	|	expr XOR expr                    { $$ = new expr_binary($1, XOR, $3); }
	|	expr OR expr                     { $$ = new expr_binary($1, OR,  $3); }
	|	expr AND expr                    { $$ = new expr_binary($1, AND, $3); }
	|	NOT expr                         { $$ = new expr_unary(NOT, $2); }
	
	|	expr '>' expr                    { $$ = new expr_binary($1, '>', $3); }
	|	expr '<' expr                    { $$ = new expr_binary($1, '<', $3); }
	|	expr '=' expr                    { $$ = new expr_binary($1, '=', $3); }
	|	expr LESSOREQUAL expr            { $$ = new expr_binary($1, LESSOREQUAL, $3); }
	|	expr GREATEROREQUAL expr         { $$ = new expr_binary($1, GREATEROREQUAL, $3); }
	|	expr NOTEQUAL expr               { $$ = new expr_binary($1, NOTEQUAL, $3); }

	|	expr '+' expr                    { $$ = new expr_binary($1, '+', $3); }
	|	expr '-' expr                    { $$ = new expr_binary($1, '-', $3); }
	|	expr MOD expr                    { $$ = new expr_binary($1, MOD, $3); }
	|	expr '\\' expr                   { $$ = new expr_binary($1, '\\', $3); }
	|	expr '/' expr                    { $$ = new expr_binary($1, '/', $3); }
	|	expr '*' expr                    { $$ = new expr_binary($1, '*', $3); }
	|	expr '^' expr                    { $$ = new expr_binary($1, '^', $3); }
	
	|       '-' expr %prec NEGATIONPRIORITY  { $$ = new expr_unary('-', $2); }
	
	|	'(' expr ')'                     { $$ = $2 }                                      
	
	|	expr_source                       { $$ = $1 }
;

numexpr:
		expr	/* FIXME - check expression type */
;
boolexpr:
		expr	/* FIXME - check expression type */
;
stringexpr:
		expr	/* FIXME - check expression type */
;

string_or_num_expr:
		expr	/* FIXME - check expression type */
;
dummy_expr:
		expr    /* discard the result */
;

int_const:
		NUMBER
;

expr_source:
		builtinfunc
	|	NUMBER
	|	STRINGCONST     { $$ = new expr_string(*$1); }
	|	var_or_const_or_functioncall
;


/****** IDENTIFIER AS A RVAL *************/
indexlist:
		indexlist ',' numexpr
			{ $1->push_back($3); $$ = $1 }
	|	numexpr
			{ $$ = new expressions($1)   }
;

var_or_const_or_function_name:
		IDENTIFIER
		{
			/* It may be a CONST, a SCALAR or a FUNCTION name. Decide which. */
			if(IsFunc(*$1))
		        {
		        	$$ = new expr_userfun(*$1);
		        }
		        else if(IsConst(*$1))
		        {
		        	/* Todo: substitute value */
		        }
		        else if(IsVar(*$1))
		        {
		        	$$ = new expr_var(*$1);
		        }
		        else
		        {
		        	YYABORT;
		        }
		}
	|	IDENTIFIER '(' subcallparams2 ')'
		{
			/* It may be an ARRAY name or a FUNCTION name. Decide which. */
			if(IsFunc(*$1))
		        {
				expr_userfun* p = new expr_userfun(*$1);
			        p->set_params($3); $3=0;
			        $$ = p;
		        }
		        else if(IsVar(*$1))
		        {
				expr_var* p = new expr_var(*$1);
				p->set_indices($3); $3=0;
				$$ = p;
		        }
		        else
		        {
		        	YYABORT;
		        }
		}
	

var_or_const_or_functioncall:
		var_or_const_or_functioncall '.' var_or_const_or_function_name
		{
			expr_var* parent = dynamic_cast<expr_var*>($1);
			expr_var* member = dynamic_cast<expr_var*>($3);
			if(!parent || !member) { YYABORT; }
			parent->set_ptr(member);
			$3 = 0;
			$$ = $1;
		}
	|	var_or_const_or_function_name
;
/************************************/


/**** IDENTIFIER AS A LVAL *********/
var_target_name:
		IDENTIFIER '(' indexlist ')'
		{
			if(IsConst(*$1) || IsSubFunc(*$1)) { YYABORT; }
			expr_var* p = new expr_var(*$1);
			p->set_indices($3); $3=0;
			$$ = p;
		}
	|	IDENTIFIER
		{
			if(IsConst(*$1) || IsSubFunc(*$1)) { YYABORT; }
			$$ = new expr_var(*$1);
		}
;
var_target:
		var_target '.' var_target_name
			{ expr_var* parent = dynamic_cast<expr_var*>($1);
		          expr_var* member = dynamic_cast<expr_var*>($3);
		          parent->set_ptr(member); $3 = 0; $$ = $1; }
	|	var_target_name
;
/************************************/


type:
		INTEGER
	|	SINGLE
	|	DOUBLE
	|	STRING
	|	LONG
	|	IDENTIFIER
		{
			if(!IsType(*$1)) { YYABORT; }
		}
;

plain_coordinate:
		'(' numexpr ',' numexpr ')'
;
coordinate:
		STEP plain_coordinate
	|	plain_coordinate
;
optional_coordinate:
		coordinate
	|	/* empty */
;

builtinfunc:
		INKEY_S                                                  { $$ = new expr_builtin_string(INKEY_S); }
	|	CSRLIN                                                   { $$ = new expr_builtin_numeric(CSRLIN); }
	|	FREEFILE                                                 { $$ = new expr_builtin_numeric(FREEFILE); }
	|	DATE_S                                                   { $$ = new expr_builtin_string(DATE_S); }
	|	TIME_S                                                   { $$ = new expr_builtin_string(TIME_S); }
	|	TIMER                                                    { $$ = new expr_builtin_numeric(TIMER); }
	|	ERL                                                      { $$ = new expr_builtin_numeric(ERL); }
	|	ERR                                                      { $$ = new expr_builtin_numeric(ERR); }
	|	RND '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(RND, $3); }
	|	RND                                                      { $$ = new expr_builtin_numeric(RND); }
	|	ASC '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(ASC, $3); }
	|	LEN '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(LEN, $3); }
	|	CVD '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(CVD, $3); }
	|	CVI '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(CVI, $3); }
	|	CVL '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(CVL, $3); }
	|	CVS '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(CVS, $3); }
	|	LTRIM_S '(' stringexpr ')'                               { $$ = new expr_builtin_string(LTRIM_S, $3); }
	|	RTRIM_S '(' stringexpr ')'                               { $$ = new expr_builtin_string(RTRIM_S, $3); }
	|	LCASE_S '(' stringexpr ')'                               { $$ = new expr_builtin_string(LCASE_S, $3); }
	|	UCASE_S '(' stringexpr ')'                               { $$ = new expr_builtin_string(UCASE_S, $3); }
	|	VAL '(' stringexpr ')'                                   { $$ = new expr_builtin_numeric(VAL, $3); }
	|	ENVIRON_S '(' string_or_num_expr ')'                     { $$ = new expr_builtin_string(ENVIRON_S, $3); }
	|	FILEATTR '(' autofilenum ',' numexpr ')'                 { $$ = new expr_builtin_numeric(FILEATTR, $3, $5); }
	|	FRE '(' string_or_num_expr ')'                           { $$ = new expr_builtin_numeric(FRE, $3); }
	|	INSTR '(' /*num*/expr ',' stringexpr ',' stringexpr ')'  { $$ = new expr_builtin_string(INSTR, $3, $5, $7); }
	|	INSTR '(' /*string*/expr ',' stringexpr ')'              { $$ = new expr_builtin_string(INSTR, $3, $5); }
	|	MID_S '(' stringexpr ',' numexpr ',' numexpr ')'         { $$ = new expr_builtin_string(MID_S, $3, $5, $7); }
	|	MID_S '(' stringexpr ',' numexpr ')'                     { $$ = new expr_builtin_string(MID_S, $3, $5); }
	|	LEFT_S '(' stringexpr ',' numexpr ')'                    { $$ = new expr_builtin_string(LEFT_S, $3, $5); }
	|	RIGHT_S '(' stringexpr ',' numexpr ')'                   { $$ = new expr_builtin_string(RIGHT_S, $3, $5); }
	|	INPUT_S '(' numexpr ',' autofilenum ')'                  { $$ = new expr_builtin_string(INPUT_S, $3, $5); }
	|	STRING_S '(' numexpr ',' string_or_num_expr ')'          { $$ = new expr_builtin_string(STRING_S, $3, $5); }
	|	POINT '(' numexpr ',' numexpr ')'                        { $$ = new expr_builtin_numeric(POINT, $3, $5); }
	|	POINT '(' numexpr ')'                                    { $$ = new expr_builtin_numeric(POINT, $3); }
	|	ABS '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(ABS, $3); }
	|	ATN '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(ATN, $3); }
	|	CDBL '(' numexpr ')'                                     { $$ = new expr_builtin_numeric(CDBL, $3); }
	|	CINT '(' numexpr ')'                                     { $$ = new expr_builtin_numeric(CINT, $3); }
	|	CHR_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(CHR_S, $3); }
	|	CLNG '(' numexpr ')'                                     { $$ = new expr_builtin_numeric(CLNG, $3); }
	|	COS '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(COS, $3); }
	|	CSNG '(' numexpr ')'                                     { $$ = new expr_builtin_numeric(CSNG, $3); }
	|	EXP '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(EXP, $3); }
	|	FIX '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(FIX, $3); }
	|	HEX_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(HEX_S, $3); }
	|	INP '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(INP, $3); }
	|	INPUT_S '(' numexpr ')'                                  { $$ = new expr_builtin_string(INPUT_S, $3); }
	|	INT '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(INT, $3); }
	|	LOC '(' autofilenum ')'                                  { $$ = new expr_builtin_numeric(LOC, $3); }
	|	LOF '(' autofilenum ')'                                  { $$ = new expr_builtin_numeric(LOF, $3); }
	|	LOG '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(LOG, $3); }
	|	LPOS '(' numexpr ')'                                     { $$ = new expr_builtin_numeric(LPOS, $3); }
	|	MKD_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(MKD_S, $3); }
	|	MKI_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(MKI_S, $3); }
	|	MKL_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(MKL_S, $3); }
	|	MKS_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(MKS_S, $3); }
	|	OCT_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(OCT_S, $3); }
	|	PEEK '(' numexpr ')'                                     { $$ = new expr_builtin_numeric(PEEK, $3); }
	|	POS '(' dummy_expr ')'                                   { $$ = new expr_builtin_numeric(POS); }
	|	PLAY '(' dummy_expr ')'                                  { $$ = new expr_builtin_numeric(PLAY); }
	|	SPC '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(SPC, $3); }
	|	SQR '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(SQR, $3); }
	|	STR_S '(' numexpr ')'                                    { $$ = new expr_builtin_string(STR_S, $3); }
	|	SPACE_S '(' numexpr ')'                                  { $$ = new expr_builtin_string(SPACE_S, $3); }
	|	TAB '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(TAB, $3); }
	|	TAN '(' numexpr ')'                                      { $$ = new expr_builtin_numeric(TAN, $3); }
	|	LBOUND '(' var_target ',' numexpr ')'                        { $$ = new expr_builtin_numeric(LBOUND, $3, $5); }
	|	UBOUND '(' var_target ',' numexpr ')'                        { $$ = new expr_builtin_numeric(UBOUND, $3, $5); }
	|	LBOUND '(' var_target ')'                                    { $$ = new expr_builtin_numeric(LBOUND, $3); }
	|	UBOUND '(' var_target ')'                                    { $$ = new expr_builtin_numeric(UBOUND, $3); }
	|	VARPTR   '(' var_target ')'                                  { $$ = new expr_builtin_numeric(VARPTR, $3); }
	|	VARPTR_S '(' var_target ')'                                  { $$ = new expr_builtin_string(VARPTR_S, $3); }
	|	VARSEG   '(' var_target ')'                                  { $$ = new expr_builtin_numeric(VARSEG, $3); }
;

%%
/* Additional C code */


int main(void)
{
    /*setbuf(stdin,NULL);*/
    
    yyStartFile(stdin, "<stdin>");
    
    yydebug=1;
    yyparse();
    return 0;
}
