#ifndef bqtQBlibGfxAPIHH
#define bqtQBlibGfxAPIHH

#include <utility>
#include <cstdio> // for size_t

#include <qblib/error.hh>

class Screen
{
public:
    virtual ~Screen() { }
    
    /* Terminal specific */
    virtual std::pair<int,int> GetCursorPos() = 0; // x, y
    virtual void SetCursorPos(const std::pair<int,int>& newpos) = 0; // x, y
    virtual void SetCursorVisibility(bool visible) = 0;
    virtual void SetCursorSize(int start, int stop) = 0;
    
    virtual void SetPages(int npages) = 0;
    virtual void SetAPage(int pageno) = 0;
    virtual void SetVPage(int pageno) = 0;
    
    virtual std::pair<int,int> GetWinSize() = 0;   // x, y
    virtual std::pair<int,int> ReadScreen(int x,int y) = 0; // chr, color
    virtual void SetWinSize(const std::pair<int,int>& newsize, bool approximate) = 0; // x, y

    virtual void OutS(const char* buffer, size_t nbytes) = 0;
    virtual void ScreenPOKE(int x,int y, int colorflag, int value) = 0;
    
    virtual void ClearEntireScreen() = 0;
    virtual void ClearText(int firstline, int lastline) = 0;
    virtual void ClearGFXport() = 0;
    
    virtual void SetColor(int fg=-1, int bg=-1, int border=-1) = 0;

    /* View print */
private:
    std::pair<int,int> ViewPrint;
public:
    const std::pair<int,int>& GetViewPrint() const { return ViewPrint; }
    void ResetViewPrint()
    {
        std::pair<int,int> limits = GetWinSize();
        ViewPrint.first  = 1;
        ViewPrint.second = limits.second;
    }
    void SetViewPrint(int from, int to)
    {
        if(from == -1 && to == -1)
        {
            from = 1;
            to = GetWinSize().second;
        }
        
        ViewPrint.first = from;
        ViewPrint.second = to;
        std::pair<int,int> pos = GetCursorPos();
        /*
        fprintf(stderr, "viewprint(%d,%d); cursor(%d,%d)\n",
            from,to, pos.first,pos.second); fflush(stderr);*/
        if(pos.second < from)      { pos.second = from; SetCursorPos(pos); }
        else if(pos.second >   to) { pos.second = to;   SetCursorPos(pos); }
    }
    void CheckViewPrint(int line)
    {
        if(line < ViewPrint.first || line > ViewPrint.second)
            B_ERROR(5); // illegal function call
    }
    
    /* CLS */
    void CLS(int type)
    {
/*
    CLS [-1] Clears either the text or graphics viewport. If a graphics
             viewport has been set (using VIEW), clears only the graphics
             viewport. Otherwise, clears the text viewport or entire screen.
    CLS 0    Clears the screen of all text and graphics.
    CLS 1    Clears the graphics viewport or the entire screen if no graphics
             viewport has been set.
    CLS 2    Clears the text viewport.
*/
        if(type == 0)
        {
            ClearEntireScreen();
            return;
        }
        bool has_gfx_viewport = false;
        if((type < 0 && !has_gfx_viewport) || type == 2)
            ClearText(ViewPrint.first, ViewPrint.second);
        else if((type < 0 && has_gfx_viewport) || type == 1)
            ClearGFXport();
    }
};

#endif
